package ru.dbotthepony.mc.otm.block.entity.tech

import it.unimi.dsi.fastutil.objects.ObjectArrayList
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import net.minecraft.core.BlockPos
import net.minecraft.core.SectionPos
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity
import ru.dbotthepony.mc.otm.block.entity.blackhole.BlackHoleBlockEntity
import ru.dbotthepony.mc.otm.config.MachinesConfig
import ru.dbotthepony.mc.otm.core.Multiblock
import ru.dbotthepony.mc.otm.core.MultiblockBuilder
import ru.dbotthepony.mc.otm.core.getBlockStateNow
import ru.dbotthepony.mc.otm.core.immutableList
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.math.plus
import ru.dbotthepony.mc.otm.core.math.times
import ru.dbotthepony.mc.otm.core.multiblockConfiguration
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.registry.MBlocks

class BlackHoleGeneratorBlockEntity(blockPos: BlockPos, blockState: BlockState) : MatteryDeviceBlockEntity(MBlockEntities.BLACK_HOLE_GENERATOR, blockPos, blockState) {
	private var multiblock: Multiblock? = null
	private var lastRange = -1

	private fun findBlackHoleRange(): Int {
		val normal = blockRotation.normal
		val level = level!!

		if (lastRange != -1) {
			val pos = blockPos + normal * lastRange

			if (level.getBlockStateNow(pos).block == MBlocks.BLACK_HOLE) {
				return lastRange
			}
		}

		for (i in 5 until 64) {
			val pos = blockPos + normal * i

			if (level.getBlockStateNow(pos).block == MBlocks.BLACK_HOLE) {
				return i
			}
		}

		return -1
	}

	override fun tick() {
		super.tick()

		val level = level!!

		if (lastRange == -1) {
			val found = findBlackHoleRange()

			if (found != -1) {
				lastRange = found
				multiblock = CONFIGURATIONS[found - 5].value.create(blockPos)
			} else {
				multiblock = null
			}
		}

		val multiblock = multiblock

		if (multiblock != null) {
			if (!multiblock.update(level, blockRotation.front)) {
				val found = findBlackHoleRange()

				if (found == -1) {
					this.multiblock = null
				} else if (found != lastRange) {
					lastRange = found
					this.multiblock = CONFIGURATIONS[found - 5].value.create(blockPos)
				}
			} else {
				val blackHole = multiblock.blockEntities(BLACK_HOLE).first()
				val matterHatches = multiblock.blockEntities(MatterHatchBlockEntity.INPUT_TAG)
				val energyHatches = multiblock.blockEntities(EnergyHatchBlockEntity.OUTPUT_TAG)

				if (matterHatches.isEmpty() || energyHatches.isEmpty()) return

				var required = MachinesConfig.BlackHoleGenerator.MATTER_RATE

				for (hatch in matterHatches) {
					required -= hatch.matter.extractMatter(required, true)
					if (required <= Decimal.ZERO) break
				}

				if (required > Decimal.ZERO) return

				var energyLeftover = blackHole.mass / MachinesConfig.BlackHoleGenerator.MASS_DIVISOR

				for (hatch in energyHatches) {
					energyLeftover -= hatch.energy.receiveEnergy(energyLeftover, true)
					if (energyLeftover <= Decimal.ZERO) break
				}

				if (energyLeftover > Decimal.ZERO) return

				required = MachinesConfig.BlackHoleGenerator.MATTER_RATE

				for (hatch in matterHatches) {
					required -= hatch.matter.extractMatter(required, false)
					if (required <= Decimal.ZERO) break
				}

				energyLeftover = blackHole.mass / MachinesConfig.BlackHoleGenerator.MASS_DIVISOR

				for (hatch in energyHatches) {
					energyLeftover -= hatch.energy.receiveEnergy(energyLeftover, false)
					if (energyLeftover <= Decimal.ZERO) break
				}

				blackHole.mass += MachinesConfig.BlackHoleGenerator.MASS_FEEDING_RATIO * MachinesConfig.BlackHoleGenerator.MATTER_RATE
			}
		}
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu? {
		return null
	}

	companion object {
		val BLACK_HOLE = MultiblockBuilder.EntityTag(BlackHoleBlockEntity::class)

		private fun points(x: Int, z: Int, result: ObjectArraySet<BlockPos>) {
			if (x == 0) {
				result.add(BlockPos(0, 0, z))
				result.add(BlockPos(0, 0, -z))
				result.add(BlockPos(z, 0, 0))
				result.add(BlockPos(-z, 0, 0))
			} else if (x == z) {
				result.add(BlockPos(x, 0, z))
				result.add(BlockPos(-x, 0, z))
				result.add(BlockPos(x, 0, -z))
				result.add(BlockPos(-x, 0, -z))
			} else if (x < z) {
				result.add(BlockPos(x, 0, z))
				result.add(BlockPos(-x, 0, z))
				result.add(BlockPos(x, 0, -z))
				result.add(BlockPos(-x, 0, -z))

				result.add(BlockPos(z, 0, x))
				result.add(BlockPos(-z, 0, x))
				result.add(BlockPos(z, 0, -x))
				result.add(BlockPos(-z, 0, -x))
			}
		}

		private fun ring(size: Int): ObjectArrayList<BlockPos> {
			val positions = ObjectArraySet<BlockPos>()
			var x = 0
			var y = size
			var p = (5 - size * 4) / 4
			points(x, y, positions)

			while (x < y) {
				x++

				if (p < 0) {
					p += x * 2 + 1
				} else {
					y--
					p += (x - y) * 2 + 1
				}

				points(x, y, positions)
			}

			val result = ObjectArrayList<BlockPos>(positions.size)

			for (pos in positions) {
				if (pos.x == 0 && pos.z + size == 0) continue
				result.add(BlockPos(pos.x, 0, pos.z + size))
			}

			return result
		}

		val CONFIGURATIONS = immutableList {
			for (i2 in 0 until 59) {
				val i = i2 + 5

				accept(lazy {
					multiblockConfiguration {
						builder.customCheck {
							val blocks = it.blocks()

							blocks.getInt(MBlocks.MATTER_INJECTOR) == 1 &&
							blocks.getInt(MBlocks.ANTIMATTER_INJECTOR) == 1 &&
							blocks.getInt(MBlocks.HIGH_ENERGY_PARTICLE_COLLECTOR) == 1
						}

						block(MBlocks.BLACK_HOLE_GENERATOR)

						relative(BlockPos(0, 0, i)) {
							block(MBlocks.BLACK_HOLE)
							tag(BLACK_HOLE)
						}

						val ring = ring(i)

						for (pos in ring) {
							val node = relative(pos)

							if ((pos.x == i || pos.x == -i) && pos.z == i) {
								node.block(MBlocks.MATTER_INJECTOR)
								node.block(MBlocks.ANTIMATTER_INJECTOR)
								node.tagBlock()
							} else if (pos.x == 0 && pos.z == i * 2) {
								node.block(MBlocks.HIGH_ENERGY_PARTICLE_COLLECTOR)
								node.tagBlock()
							} else if (pos.x == i || pos.x == -i || pos.z == i * 2 || pos.z == 0) {
								node.or {
									block(MBlocks.ENERGY_OUTPUT_HATCH)
									tag(EnergyHatchBlockEntity.OUTPUT_TAG)
								}

								node.or {
									block(MBlocks.MATTER_INPUT_HATCH)
									tag(MatterHatchBlockEntity.INPUT_TAG)
								}

								node.block(MBlocks.MULTIBLOCK_STRUCTURE)
							} else {
								node.block(MBlocks.MULTIBLOCK_STRUCTURE)
							}
						}
					}
				})
			}
		}
	}
}
